{-
   Area3D.hs
     - this module contains all the rendering functions needed to
       create 3D object of the Mandelbrot set
     - it uses function evaluateAt to do all the mathematical computations

   Author:       Martin Furman (s0830838)
   Organisation: University of Edinburgh
   Email:        mrt.furman@gmail.com
   Created:      22/11/2008
-}

module Area3D where

import Types
import Evaluator
import GlobalValues

import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT

-- create the Mandelbrot set as the array of interconnected quadrangles
-- in patched version "preservingMatrix" function is left out as it
-- requires the double buffered environment
renderArea3D :: Area3D -> IO ()
renderArea3D area = preservingMatrix $ do
   renderPrimitive Quads $ createArea res oX oY sc
    where
    oX = offsetX area
    oY = offsetY area
    sc = viewScale area
    qual = quality area
    -- resolution: see GlobalValues module 
    res | qual == LowQual = mLowResolution
        | otherwise       = mHighResolution
-- the end: renderArea3D



-- generates the final area that is to be rendered on the screen
createArea :: GLsizei -> GLfloat -> GLfloat -> GLfloat -> IO ()
createArea res oX oY sc = convertPoints mArea
   where
   -- creates the list of list of all the xy and ys from a range
   -- and evaluate them together with offsets and scale
   mArea :: MArea
   mArea = [ [ evaluateAt x y oX oY sc | x <- range] | y <- range]
   
   range :: [GLfloat]
   range = [-1.0,(-1.0+step)..1.0]

   -- step calculated from resolution
   step :: GLfloat
   step = 1.0 / (fromIntegral res)
-- the end: createArea



-- coversion of simple points to points ready to be used with
-- quadrangle rendering
-- example: 2x2 area
--   (top left)    (top right)
--   (bottom left) (bottom right)
-- will be rendered counter clockwise
--   (top left)    <-   (top right)
--                          ^
--   (bottom left)  ->  (bottom right)
-- eventually, for greater areas top left point becomes
-- the bottom left point of the next row...
convertPoints :: MArea -> IO ()
convertPoints area = mergePoints area $ leftBottomPoints area
-- the end: convertPoints



-- does the conversion for every bottom left point of every single
-- rectangle (see convertPoint function description)
mergePoints :: MArea -> [(Int, Int)] -> IO ()
mergePoints area (p:[]) = mergePoint area p
mergePoints area (p:ps) = mergePoint area p >> mergePoints area (ps)
-- the end: mergePoints



-- does the actual rendering (see convertPoint function description)
-- every point has its color set as well
-- the patched version omits the preservingMatrix function as it does not
-- work properly without double buffered environment
mergePoint :: MArea -> (Int, Int) -> IO ()
mergePoint area p = preservingMatrix $ do
  color  $ colorOf  $ getPoint area p
  vertex $ vertexOf $ getPoint area p
  color  $ colorOf  $ pointNext area p
  vertex $ vertexOf $ pointNext area p
  color  $ colorOf  $ pointDiagonal area p
  vertex $ vertexOf $ pointDiagonal area p
  color  $ colorOf  $ pointAbove area p
  vertex $ vertexOf $ pointAbove area p
    where
    vertexOf :: MPoint -> Vertex3 GLfloat
    vertexOf ((x,y,z), _ ) = Vertex3 x y z

    colorOf :: MPoint -> Color3 GLfloat
    colorOf (_,x) = x

    pointAbove, pointDiagonal, pointNext :: MArea -> (Int, Int) -> MPoint
    pointAbove area (x,y) = getPoint area (x,y+1)
    pointDiagonal area (x,y) = getPoint area (x+1,y+1)
    pointNext area (x,y) = getPoint area (x+1,y)

    getPoint :: MArea -> (Int, Int) -> MPoint
    getPoint area (x, y) = (area !! y) !! x
-- the end: mergePoint



-- (x,y) pairs for every point except of top row and last column,
-- as these do not have any other point above/next to/diagonally
leftBottomPoints :: MArea -> [(Int, Int)]
leftBottomPoints area = [ (x,y) | x <- [0..xs], y <- [0..ys] ]
  where
  xs = (length (area !! 0)) - 2
  ys = (length area) - 2
-- the end: leftBottomPoints

