{-
   Display.hs
     - this module encapsulate final visual output functions
       and user interaction functions

   Author:       Martin Furman (s0830838)
   Organisation: University of Edinburgh
   Email:        mrt.furman@gmail.com
   Created:      22/11/2008
-}

module Display where

import Types
import Area3D
import GlobalValues

import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT
import Data.IORef



display :: IORef Area3D -> IORef (GLmatrix GLdouble) -> IO ()
display area3D rotMatrix  = do
  prevMatrix <- get rotMatrix
  area <- get area3D

  -- FIXIT: re-drawing of high quality picture is disabled due to the long time
  --        it takes to do so, but there is a case, when the window needs to
  --        re-drawn if it was hidden/iconified/minimised etc. 
  if not (isChanged area) && quality area == HighQual then return ()
                                                      else clear [ColorBuffer]


  let rotation = (viewRotation area)
      rotX = (rotationX rotation)
      rotY = (rotationY rotation)
      rotZ = (rotationZ rotation)

  loadIdentity
  let xaxis = (Vector3 1 0 0)
      yaxis = (Vector3 0 1 0)
      zaxis = (Vector3 0 0 1)

  rotate (rotX) xaxis
  rotate (rotY) yaxis
  rotate (rotZ) zaxis

  multMatrix prevMatrix
  currentMatrix <- get ((matrix $ Just $ Modelview 0) :: StateVar (GLmatrix GLdouble))
  rotMatrix $= currentMatrix

  -- FIXIT: re-drawing of high quality picture is disabled due to the long time
  --        it takes to do so, but there is a case, when the window needs to
  --        re-drawn if it was hidden/iconified/minimised etc. 
  -- in patched version swapBuffers is replaced by flush
  if not (isChanged area) && quality area == HighQual then return ()
                                                      else renderArea3D area >> flush -- swapBuffers
  
  -- once the window's re-drawn, set the isChanged flat to false and zero the rotation
  area3D $= area {isChanged = False,
                  viewRotation = rotation {rotationX = 0.0, rotationZ = 0.0}}
-- the end: display



-- keboard interation - this function is called whenever mouse or keybord produce any event
keyboardMouse area3D rotMatrix key state modifiers _ = do
  area <- get area3D
  -- update area3D (IORef Area3D) with new area parametres
  let newArea = keyAction area key state modifiers
  area3D $= newArea
  -- once parapeters changed, re-draw window with new parameters
  if isChanged newArea then display area3D rotMatrix else return ()
    where
    -- following two lines represents key 'R' pressed, either with, or without Shift
    keyAction area (Char 'r') Down _ = area {quality = HighQual, viewRotation = emptyRotation, isChanged = True}
    keyAction area (Char 'R') Down _ = area {quality = HighQual, viewRotation = emptyRotation, isChanged = True}

    -- 9 predefined pictures for keys 1 to 9
    keyAction area (Char '1') Down _ = area {quality = LowQual, offsetX = (-0.13), offsetY = 0.97,
                                             viewScale = 0.16, isChanged = True}
    keyAction area (Char '2') Down _ = area {quality = LowQual, offsetX = (-0.92), offsetY = 0.25,
                                             viewScale = 0.16, isChanged = True}
    keyAction area (Char '3') Down _ = area {quality = LowQual, offsetX = (-0.925), offsetY = 0.3,
                                             viewScale = 0.03, isChanged = True}
    keyAction area (Char '4') Down _ = area {quality = LowQual, offsetX = (-0.6503), offsetY = 0.4495,
                                             viewScale = 0.002, isChanged = True}
    keyAction area (Char '5') Down _ = area {quality = LowQual, offsetX = (-0.7504), offsetY = 0.025,
                                             viewScale = 0.001, isChanged = True}
    keyAction area (Char '6') Down _ = area {quality = LowQual, offsetX = (-0.74993), offsetY = 0.02504,
                                             viewScale = 0.00006, isChanged = True}
    keyAction area (Char '7') Down _ = area {quality = LowQual, offsetX = (-0.749475), offsetY = 0.0384,
                                             viewScale = 0.0001, isChanged = True}
    keyAction area (Char '8') Down _ = area {quality = LowQual, offsetX = (-0.746), offsetY = 0.184,
                                             viewScale = 0.003, isChanged = True}
    keyAction area (Char '9') Down _ = area {quality = LowQual, offsetX = 0.26, offsetY = 0.0018,
                                             viewScale = 0.0003, isChanged = True}
    -- key 0 represents the default view
    keyAction area (Char '0') Down _ = area {quality = LowQual, offsetX = (-0.7), offsetY = 0.0,
                                             viewScale = 1.5, isChanged = True}
    -- if none of keys above, then expect arrow keys
    --    a) if shift key's down, progress to rotation
    --    b) otherwise progress to movement
    keyAction area key Down (Modifiers {shift = Down, ctrl = Up, alt = Up}) = keyRotate area key
    keyAction area key Down _ = keyMove area key
    -- every other keyAction (e.g. when key status is changing from Down to Up makes no change to the area
    keyAction area _ _ _ = area

    -- rotation functions
    keyRotate area (SpecialKey KeyUp) = area {quality = LowQual, isChanged = True,
                                              viewRotation = newUDRotation 1.0 (viewRotation area) }
    keyRotate area (SpecialKey KeyDown) = area {quality = LowQual, isChanged = True,
                                                viewRotation = newUDRotation (-1.0) (viewRotation area)}
    keyRotate area (SpecialKey KeyLeft) = area {quality = LowQual, isChanged = True,
                                                viewRotation = newLRRotation 1.0 (viewRotation area)}
    keyRotate area (SpecialKey KeyRight) = area {quality = LowQual, isChanged = True,
                                                 viewRotation = newLRRotation (-1.0) (viewRotation area)}
    -- !!! scale function
    -- confusing code, but under cover of following rotation function is acually scale function
    keyRotate area (Char ' ') = area {quality = LowQual, viewRotation = emptyRotation,
                                      isChanged = True, viewScale = (1.0 + mScaleStep )* (viewScale area)}
    -- if other key than those above, then do not change area
    keyRotate area _ = area

    -- aux. rotation up and down function
    newUDRotation i rot = rot {rotationX =  mRotationStep * i,
			     rotationZ = 0.0}
    -- aux. rotation left and right function
    newLRRotation i rot = rot {rotationX = 0.0,
                               rotationZ = i * mRotationStep}
    -- aux. function returning empty rotation
    emptyRotation = MRotation {rotationX = 0.0, rotationY = 0.0, rotationZ = 0.0}
    
    -- move functions
    keyMove area (SpecialKey KeyUp) = area {quality = LowQual, viewRotation = emptyRotation, isChanged = True,
                                            offsetY = mMoveStep * (viewScale area) + (offsetY area)}
    keyMove area (SpecialKey KeyDown) = area {quality = LowQual, viewRotation = emptyRotation, isChanged = True,
                                            offsetY = (offsetY area) - mMoveStep * (viewScale area)}
    keyMove area (SpecialKey KeyLeft) = area {quality = LowQual, viewRotation = emptyRotation, isChanged = True,
                                            offsetX = (offsetX area) - mMoveStep * (viewScale area)}
    keyMove area (SpecialKey KeyRight) = area {quality = LowQual, viewRotation = emptyRotation, isChanged = True,
                                            offsetX = mMoveStep * (viewScale area) + (offsetX area)}
    -- !!! scale function
    -- confusing code, but under cover of following move function is actually scale function
    keyMove area (Char ' ') = area {quality = LowQual, viewRotation = emptyRotation, isChanged = True,
                                            viewScale = (viewScale area) * (1.0 - mScaleStep)}
    -- if none from above, do not change the area
    keyMove area _ = area
-- the end: keyboardMouse

