{-
   Evaluator.hs
     - main equations that determine the look of final picture (in this case
       Mandelbrot set) are stored in this module

   Author:       Martin Furman (s0830838)
   Organisation: University of Edinburgh
   Email:        mrt.furman@gmail.com
   Created:      22/11/2008
-}

module Evaluator where

import Types
import Spectrum
import GlobalValues

import Graphics.Rendering.OpenGL
import Graphics.UI.GLUT

-- just to make it easier to write things down i introduces addition
-- type MResult witch contains the lift and the color of the pixel
type MResult = (GLfloat, Color3 GLfloat)



-- set the known variable of the MPoint and let the getResult function
-- compute the others
evaluateAt :: GLfloat -> GLfloat -> GLfloat -> GLfloat -> GLfloat -> MPoint
evaluateAt x y oX oY scale = ((x,y,lift), color)
   where
   -- offsets divided by scale so they are of the same "weight" as x and y
   -- thus pointing to the same direction regardless the scale
   (lift,color) = getResult scale (x + oX / scale) (y + oY / scale)
-- the end: evaluateAt



-- getResult makes the transformations from number of steps to
-- lift and color (using getSpectrum color in Spectrum module)
getResult :: Float -> Float -> Float -> MResult
getResult scale x y = (getLift, getColor)
  where
  getLift = mMaxLift / mColorDepth * (fromIntegral (toInteger steps)) 
  getColor = getSpectrum ((fromIntegral (toInteger(steps))) / (mColorDepth))
  
  steps = countSteps (x * scale) (y * scale)
-- the end: getResult



-- definition of Mandelbrot set:
--    1) take complex number C (x,y) and Z (zX,zY)
--    2) check whether the length of Z is greater than 2 (lngth)
--       and whether we have done maximal allowed number of iterations
--         a) if length greater than 2 return number of iterations
--         b) if reached maximal allowed number of iterations return 0 (black)
--         c) if neither continue to 3)
--    3) calculate new complex number Z = Z^2 + C (newZX,newZY)
--    4) go back to 2)
countSteps x y = counter x y 0 0 0 mSteps
  where
  counter x y zX zY i max 
     | lngth <= 2.0 && i < max = counter x y newZX newZY (i+1) max
     | lngth > 2.0             = i
     | i == max                = 0
       where
       lngth = sqrt (zX*zX + zY*zY)
       newZX = zX*zX - zY*zY + x
       newZY = 2*zX*zY + y
-- the end: countSteps

